/* Copyright (C) 2006 P.L. Lucas
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA. 
 */

#include "octave_connection.h"
#include "config.h"
#include <QThread>
#include <QTextCursor>
#include <QMessageBox>

OctaveConnection::OctaveConnection(QObject * parent ):QProcess(parent)
{
	octave_path="";
	//text=NULL;
	error_buffer.open(QBuffer::ReadOnly);
	QObject::connect(this, SIGNAL(readyReadStandardError()), this, SLOT(octaveErrorOutputSlot()));
	
	QObject::connect(this, SIGNAL(readyReadStandardOutput()), this, SLOT(octaveOutputSlot()));
	
	QObject::connect(this, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(octaveFinished(int, QProcess::ExitStatus)) );
}

void OctaveConnection::setOctavePath(const QString path)
{
	octave_path=path;
}

QString OctaveConnection::getOctavePath()
{
	return octave_path;
}

// void OctaveConnection::setTextEdit(QTextEdit *text)
// {
// 	this->text=text;
// 	QObject::connect(this, SIGNAL(readyReadStandardError()), this, SLOT(octaveErrorOutputSlot()));
// 	QObject::connect(this, SIGNAL(readyReadStandardOutput()), this, SLOT(octaveOutputSlot()));
// }

void OctaveConnection::startOctave(bool quiet)
{
  QString extra_args = get_config("octave_arguments");

	#if defined Q_OS_WIN32
		FILE *in=popen(QString(octave_path+"cscript //NoLogo octave_pos.js").toLocal8Bit().data(),"r");
		if(in==NULL) return;
		char command[1024];
		if( fgets(command,1024,in)==NULL ) return;
		pclose(in);
		if(quiet)
			start(command+QString(" octave_quiet.sh")+extra_args );
		else
			start(command+QString(" octave.sh")+extra_args );
		waitForStarted (-1);
		if(!quiet) command_enter("PS1=\'\';PS2=\'\';disp(\"Starting octave ...\");disp(\"Octave ready.\");",false);
	#else
		if(quiet)
			start(octave_path+"octave --eval \"PS1(\'\');PS2(\'\');\" --persist -q -i"+extra_args);
		else
			start(octave_path+"octave --eval \"PS1(\'\');PS2(\'\');\" --persist -i"+extra_args);

	#endif
  //start(octave_path+"octave -i");
}

void OctaveConnection::command_enter (const QString &command, bool show)
{
	if(show)
	{
		/*
		QTextCursor cursor=text->textCursor();
		cursor.movePosition(QTextCursor::End);
		text->setTextCursor( cursor );
		text->setFontWeight ( QFont::Bold );
		text->setTextColor( Qt::black );
		text->insertPlainText(command+"\n");
		text->setFontWeight ( QFont::Normal );
		*/
		emit command_ready(">> "+command+"\n");
	}
	write((command+"\n").toLocal8Bit() );
}


void OctaveConnection::octaveOutputSlot()
{
	QString buffer=readAllStandardOutput();
	QRegExp regexp;
	QThread::currentThread ()->wait (100);

	emit output_ready(buffer);

	// Partir en líneas
	QStringList lines = buffer.split("\n", QString::SkipEmptyParts);

	if(lines.empty())
	  lines.push_back("");

	// Al primer elemento se le debe concatenar el buffer de línea
	lines.first() = line_buffer + lines.first();
	line_buffer.clear();

	// En caso de que line_buffer sea debug> u octave:X>, no guardarlo
	if(line_buffer == "debug> ")
	  lines.push_back(line_buffer);

	// Procesar las líneas
	for(QStringList::iterator i = lines.begin();
	    i != lines.end(); i++)
	{
	  // Debug
	  if(i->startsWith("debug> "))
	  {
	    i->remove(0, 7);
	    debugging = true;
	    emit debug(lineno, colno);
	  }

	  // octave:X>
	  regexp.setPattern("^octave:[0-9]+> ");
	  if(regexp.indexIn(*i) > -1)
	  {
	    i->remove(0, regexp.matchedLength());

	    if(debugging)
	    {
	      debugging = false;
	      emit endDebug();
	    }
	  }

	  // Línea y columna cuando va a parar
	  regexp.setPattern("^([a-zA-Z_]+\\w+): line [0-9]+, column [0-9]+$");
	  if(regexp.exactMatch(*i))
	  {
	    int start;

	    // Extraer la línea
	    regexp.setPattern("line [0-9]+");
	    start = regexp.indexIn(*i);

	    if(start > -1)
	      lineno = i->mid(start + 5, regexp.matchedLength() - 5).toInt();

	    // Extraer la columna
	    regexp.setPattern("column [0-9]+");
	    start = regexp.indexIn(*i);
	    if(start > -1)
	      colno = i->mid(start + 7, regexp.matchedLength() - 7).toInt();
	  }
	}

	// Meter en el buffer de línea el último elemento si la cadena
	// no terminaba en \n
	if(!buffer.endsWith("\n"))
	  line_buffer = lines.last();
}

void OctaveConnection::octaveErrorOutputSlot()
{
	//QTextCursor cursor=text->textCursor();
	//cursor.movePosition(QTextCursor::End);
	//text->setTextColor( Qt::red );
	//text->setFontWeight ( QFont::Normal );
	//text->insertPlainText(buffer);
	//text->setTextCursor( cursor );
	
	//QString buffer=readAllStandardError();
	
	QString buffer=error_buffer.data();
	error_buffer.close();
	error_buffer.setData( buffer.toUtf8()+readAllStandardError());
	error_buffer.open(QBuffer::ReadOnly);
	usleep(200);
	
	QString line, error("");
	do
	{
		line=error_buffer.readLine();
		if(line.endsWith("\n") )
		{
			if(line.startsWith("~~")) emit line_ready(line);
			else error+=line;
			continue;
		}
		else
		{
			line=line+error_buffer.readAll();
			error_buffer.close();
			error_buffer.setData(line.toUtf8());
			error_buffer.open(QBuffer::ReadOnly);
			break;
		}
		emit line_ready(line);
	}
	while(error_buffer.bytesAvailable());
	emit error_ready(error);
	if(!error_buffer.bytesAvailable())
	{
		error_buffer.close();
		error_buffer.setData(QString("").toUtf8());
		error_buffer.open(QBuffer::ReadOnly);
	}
	
	//Se procesa el buffer para buscar las l�eas diponibles
	//QStringList lines=buffer.split("\n");
	//QStringList::const_iterator i;
	//for (i = lines.constBegin(); i != lines.constEnd();)
	//{
	//	QString line=*i;
	//	++i;
	//	if(i==lines.constEnd() && !buffer.endsWith("\n"))
	//		emit error_ready(line);
	//	else
	//		emit error_ready(line+"\n");
	//	emit line_ready(line);
	//}
}

void OctaveConnection::octaveFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
	if(exitStatus == QProcess::CrashExit)
	{
		QMessageBox::critical(NULL,tr("Octave crashed"), tr("Octave has crashed. Restarting..."));
	}
	else
	{
		QMessageBox::critical(NULL,tr("Octave finished"), tr("Octave has finished. Restarting..."));
	}
	//Relanzar
	startOctave();
}

